import enum
import json
import logging
from datetime import date, datetime, timedelta

from pydantic import ValidationError
from pydantic.main import MetaModel

from sbus.models import Response

from .exceptions import BadRequestError, InternalServerError, SerializationError

logger = logging.getLogger(__name__)


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code."""

    if isinstance(obj, (datetime, date)):
        serial = obj.isoformat()

    elif issubclass(obj.__class__, enum.Enum):
        serial = obj.value

    elif isinstance(obj, timedelta):
        serial = str(obj)

    else:
        raise SerializationError(f'Type not serializable {type(obj)} in {obj}')

    return serial


class JSONSerializer:
    def serialize(self, data: Response) -> bytes:
        try:
            serialized = json.dumps(
                data.dict(),
                ensure_ascii=False,
                default=json_serial).encode()
        except Exception as error:
            logger.exception('Can\'t serialization message: %s. Because of %s', data, str(error))
            raise SerializationError from error
        return serialized

    def deserialize(self, msg, model: MetaModel) -> MetaModel:
        if not isinstance(model, MetaModel):
            raise InternalServerError('Use pydantic for describing expected request type.')

        body_txt = msg.decode('utf-8') if hasattr(msg, 'decode') else msg
        try:
            deserialized = json.loads(body_txt)
        except json.JSONDecodeError as error:
            logger.exception('Can\'t deserialize message: %s. Because of %s',
                             body_txt, str(error))
            raise SerializationError from error

        try:
            return model(**deserialized.get('body'))
        except ValidationError as error:
            logger.exception('Invalid message: %s. Because of %s', body_txt, str(error))
            raise BadRequestError(str(error)) from error



import asyncio
from unittest import mock
from unittest.mock import sentinel


def make_mocked_coro(return_value=sentinel, raise_exception=sentinel):
    """Creates a coroutine mock."""
    @asyncio.coroutine
    def mock_coro(*args, **kwargs):
        if raise_exception is not sentinel:
            raise raise_exception
        return return_value

    return mock.Mock(wraps=mock_coro)



import typing

from pydantic import BaseModel, Required


class Headers:
    correlation_id = 'correlation-id'

    message_id = 'message-id'

    client_message_id = 'client-message-id'

    retry_attempts_max = 'retry-attempts-max'

    retry_attempt_nr = 'retry-attempts-nr'

    expired_at = 'expired-at'

    timeout = 'timeout'

    routing_key = 'routing-key'


class Response(BaseModel):
    status: int = 200
    body: typing.Any = None


class ResponseBody(BaseModel):
    error: str = None
    message: str = None


class Context(BaseModel):
    routing_key: str = Required

    message_id: str = None



import asyncio
import logging
import os
import sys

import pytest

from sbus.router import Router
from sbus.subscribers import QueueSubscriber
from sbus.transports import AMQPTransport
from tests.utils import make_mocked_coro

logger = logging.getLogger(__name__)


class Context:

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop
        self._on_shutdown = []

    def enable_logging(self, level=logging.DEBUG):
        current_level = logging.getLogger().getEffectiveLevel()
        logging.basicConfig(level=level, stream=sys.stderr)

        def _disable_logging():
            logging.basicConfig(level=current_level)

        self.add_on_shutdown(_disable_logging)

    def create_queue(self, size=0, timeout=None):
        queue = asyncio.Queue(maxsize=size, loop=self.loop)

        def _close_queue():
            if not queue._getters:
                return None

            # there are some waiters for messages,
            for getter in queue._getters:
                getter.set_result(None)

            if timeout:
                pytest.fail(f'Queue timeout "{timeout}" seconds has been reached')
            else:
                raise pytest.fail('Queue is not released until test shutdown')

        if timeout:
            self.loop.call_later(timeout, _close_queue)
        else:
            self.add_on_shutdown(_close_queue)

        return queue

    def create_event(self, timeout=None):
        event = asyncio.Event(loop=self.loop)

        def _set_event():
            if event.is_set():
                return
            event.set()
            if timeout:
                pytest.fail(f'Timeout "{timeout}" seconds has been reached')
            else:
                pytest.fail('Event was not set until test shutdown')

        if timeout:
            self.loop.call_later(timeout, _set_event)
        else:
            self.add_on_shutdown(_set_event)
        return event

    def add_on_shutdown(self, callback):
        self._on_shutdown.append(asyncio.coroutine(callback)())

    def shutdown(self):
        if self._on_shutdown:
            self.loop.run_until_complete(asyncio.gather(*self._on_shutdown))

        tasks = asyncio.Task.all_tasks(self.loop)
        if tasks:
            self.loop.run_until_complete(asyncio.wait(tasks, loop=self.loop, timeout=1))


@pytest.fixture()
def ctx(event_loop):
    context = Context(event_loop)
    if os.getenv('LOG'):
        context.enable_logging(logging.DEBUG)
    yield context
    context.shutdown()


@pytest.fixture
def connection_parameters():
    return {
        'login': os.environ.get('RABBITMQ_DEFAULT_USER', 'guest'),
        'password': os.environ.get('RABBITMQ_DEFAULT_PASS', 'guest'),
        'host': os.environ.get('RABBITMQ_HOST', 'localhost'),
        'port': '5672',
        'virtualhost': os.environ.get('RABBITMQ_DEFAULT_VHOST', '/'),
    }


@pytest.fixture
def exchange_name():
    return os.environ.get('RABBITMQ_EXCHANGE', 'sbus')


@pytest.fixture
def exchange_type():
    return 'direct'


@pytest.fixture
def service_prefix():
    return 'my_test_app'


@pytest.fixture
def queue_name(service_prefix):
    return f'{service_prefix}.queue.test'


@pytest.fixture(scope='session')
def test_topic():
    return 'auth.test'


@pytest.fixture
def another_topic():
    return 'auth.test_double'


@pytest.fixture
def prefetch_count():
    return 1


@pytest.fixture
def worker_name():
    return 'test_worker'


@pytest.fixture
def wrong_worker_name():
    return 'no_test_worker'


@pytest.fixture
def routes(queue_name, test_topic, prefetch_count, worker_name, wrong_worker_name):
    return {
        worker_name: {
            'routing_keys': (test_topic,),
            'queue_name': queue_name,
            'prefetch_count': prefetch_count,
        },
        wrong_worker_name: {
            'routing_keys': (test_topic,),
            'queue_name': queue_name,
            'prefetch_count': prefetch_count,
        }
    }


@pytest.fixture
def subscriber(worker_name, router, routes):
    return QueueSubscriber(router, routes[worker_name])


@pytest.fixture()
def echo_subscriber(test_topic, worker_name, routes):
    _router = Router()

    async def echo(data, **kwargs):
        return data

    _router.register(test_topic)(echo)

    return QueueSubscriber(_router, routes[worker_name])


@pytest.fixture(scope='session')
def fake_route():
    return make_mocked_coro(return_value='test')


@pytest.fixture(scope='session', autouse=True)
def router(test_topic, fake_route):
    router_ = Router()
    router_.register(test_topic)(fake_route)
    return router_


@pytest.fixture
def amqp_client(event_loop, connection_parameters, exchange_name, exchange_type, worker_name):
    client = AMQPTransport(connection_parameters, exchange_name, exchange_type, worker_name,
                           loop=event_loop)
    event_loop.run_until_complete(client.connect())
    event_loop.run_until_complete(client.declare_exchange(exchange_name, durable=False))
    yield client
    event_loop.run_until_complete(client.destroy())



import asyncio
import logging
import typing

from .subscribers import AbstractSubscriber
from .transports.base import AbstractTransport
from .utils import create_transport

logger = logging.getLogger(__name__)


class Sbus:

    def __init__(self, transport: AbstractTransport, loop=None):
        """
        :param transport: broker transport
        :param loop: event loop instance
        """
        self.transport = transport
        self.subscribers = set()
        self.loop = loop or asyncio.get_event_loop()

    async def connect(self):
        await self.transport.connect()

    @property
    def connected(self):
        return self.transport.connected

    async def command(self, message, routing_key):
        """Publish message with chosen transport

        :param message: message for publish
        :param routing_key: routing key for message
        """
        await self.transport.command(message, routing_key)

    async def request(self, message, routing_key, response_class):
        """RPC call with data and routing key

        :param message:  message for publish
        :param routing_key: routing key for message
        :param response_class: response class
        """
        return await self.transport.request(message, routing_key, response_class)

    async def on(self, subscriber: AbstractSubscriber):
        """Subscribe on queue."""
        if not self.connected:
            return

        if subscriber in self.subscribers:
            logger.warning('Duplicate subscriber %s', subscriber)
            return

        await self.transport.on(subscriber)
        self.subscribers.add(subscriber)

    async def stop(self):
        await self.transport.destroy()
        self.subscribers.clear()

    def clone_subscribers(self):
        return self.subscribers.copy()

    @classmethod
    def create_factory(cls, config, loop):
        loop = loop or asyncio.get_event_loop()

        def _factory():
            transport = create_transport(config, loop=loop)
            return cls(transport, loop)

        return _factory


class BackOff:
    """Infinite backoff generation"""

    def __init__(self, min_timeout=1, max_timeout=15):
        self.min_timeout = min_timeout
        self.max_timeout = max_timeout
        self.retry_count = 0

    def get_timeout(self):
        self.retry_count += 1
        timeout = self.min_timeout * self.retry_count
        return timeout if timeout < self.max_timeout else self.max_timeout

    def reset(self):
        self.retry_count = 0


class SbusWatchDog:
    """Class for control Sbus health and status.

    This class should be used if you want survive through broker downs.
    """
    def __init__(self, service_bus_factory: typing.Callable, *, backoff: BackOff=None):
        """
        :param service_bus_factory: callable, must return Sbus instance
        :param backoff: BackOff instance to handle reconnection timeouts
        """
        self._service_bus_factory = service_bus_factory
        self._lock = asyncio.Lock()
        self._retry_attempts = 0
        self._backoff = backoff or BackOff()
        self._service_bus: Sbus = self._service_bus_factory()

    @property
    def sbus(self) -> Sbus:
        return self._service_bus

    async def repair(self):

        if self._lock.locked():
            logger.warning('Service bus repairing is in progress now')
            return

        with await self._lock:
            await self._repair()

    async def _repair(self):
        logger.error('Closed connection detected. Trying to reconnect to service bus.')
        backoff_timeout = self._backoff.get_timeout()

        logger.info('Reconnection attempts=%s, next in %s sec',
                    self._backoff.retry_count, backoff_timeout)

        await asyncio.sleep(backoff_timeout)

        current_service_bus: Sbus = self._service_bus
        loop = current_service_bus.loop

        subscribers = current_service_bus.clone_subscribers()

        try:
            await current_service_bus.stop()
            del current_service_bus
        except Exception:
            logger.exception('Exception occurred when stopping service bus, but may be ignored')

        try:
            sbus = self._service_bus_factory()

            await sbus.connect()
            await asyncio.gather(
                *[sbus.on(subscriber) for subscriber in subscribers],
                loop=sbus.loop
            )

            self._service_bus = sbus
            self._backoff.reset()
        except Exception:
            # several exceptions possible here, but we may safely ignore them
            logger.exception('Exception occurred when reconnecting, but may be ignored')
            loop.create_task(self.repair())
            return False

        logger.info('Repair is completed successfully')
        return True



import inspect
import logging
import typing
import uuid
from abc import ABCMeta, abstractmethod

from sbus.models import Response
from sbus.serializer import JSONSerializer

from .router import Router

logger = logging.getLogger(__name__)


class AbstractSubscriber(metaclass=ABCMeta):
    """Handle and process queue messages."""

    def __init__(self, routing: Router, subscriber_config: typing.Dict,
                 **handler_kwargs):
        assert subscriber_config, 'Subscriber configuration should be provided'
        self.config = subscriber_config
        self.routing = routing
        self.handler_kwargs = handler_kwargs

    @abstractmethod
    async def on_message(self, data, routing_key, context, serializer):
        pass

    @property
    @abstractmethod
    def consumer_tag(self):
        pass

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def routing_keys(self):
        pass

    @property
    @abstractmethod
    def durable(self):
        pass

    @property
    @abstractmethod
    def prefetch_count(self):
        pass

    def __str__(self):
        return f'<Subscriber: name={self.name} durable={self.durable}>'

    def __repr__(self):
        return self.__str__()


class QueueSubscriber(AbstractSubscriber):

    def consumer_tag(self):
        return 'consumer_tag.{}.{}'.format(self.name, uuid.uuid4().hex)

    @property
    def name(self) -> str:
        return self.config.get('queue_name', 'undefined')

    @property
    def durable(self) -> str:
        return self.config.get('durable', True)

    @property
    def prefetch_count(self) -> str:
        return self.config.get('prefetch_count', 0)

    @property
    def routing_keys(self):
        return self.routing.routing.keys()

    async def on_message(self, data: bytes, routing_key: typing.AnyStr,
                         context: typing.Dict, serializer: JSONSerializer):
        handler = self.routing.get_handler(routing_key)

        request_body = serializer.deserialize(data, self._get_request_model(handler))
        response = await handler(request_body, context, **self.handler_kwargs)
        return serializer.serialize(Response(body=response))

    @staticmethod
    def _get_request_model(handler):
        return inspect.signature(handler).parameters['task'].annotation



from .transports import SUPPORTED_TRANSPORTS


def my_import(name):
    components = name.split('.')
    mod = __import__('.'.join(components[:-1]))
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def create_transport(config, loop):
    """Create transport from configured settings.

        Will definitely fail if there is no meaningful configuration provided.
        Example of such is:
            {
                'transports': {
                    'ampq': {
                        'connection_parameters': {
                            'login': 'guest',
                            'password': '',
                            'host': 'localhost',
                            'port': 5672,
                            'virtualhost': 'sbus',
                        },
                        'exchange_name': 'sbus',
                        'exchange_type': 'topic',
                    },
                },
                'active_transport': 'amqp',
            }
        """
    active_transport = config['active_transport']

    assert active_transport in SUPPORTED_TRANSPORTS, f'Unsupported transport "{active_transport}"'

    active_transport_setup = config['transports'][active_transport]

    transport_class = SUPPORTED_TRANSPORTS[active_transport]
    if 'serializer' in active_transport_setup:
        active_transport_setup['serializer'] = my_import(
            active_transport_setup.get('serializer')
        )()
    transport = transport_class(**active_transport_setup, loop=loop)

    return transport



"""
Use this module members to connect to RabbitMQ instance.
"""

import asyncio
import logging
import math
import time
import typing  # noqa
from uuid import uuid4
from weakref import WeakValueDictionary

import aioamqp
from aioamqp import AmqpClosedConnection, ChannelClosed
from aioamqp.channel import Channel  # noqa
from aioamqp.envelope import Envelope
from aioamqp.properties import Properties
from aioamqp.protocol import CLOSED

from .. import exceptions
from ..models import Context, Headers, Response, ResponseBody
from ..serializer import JSONSerializer
from ..subscribers import AbstractSubscriber
from .base import AbstractTransport

logger = logging.getLogger(__name__)


def smart_int(value: str, default=None) -> typing.Optional[int]:
    """Cast str to int or return default value"""

    try:
        return int(float(value))
    except (ValueError, TypeError, OverflowError):
        return default


class AMQPTransport(AbstractTransport):
    """
    Handy implementation of the asynchronous AMQP client.
    Useful for building both publishers and consumers.
    """

    def __init__(
            self,
            connection_parameters,
            exchange_name,
            exchange_type,
            channel_name,
            connection_timeout=3,
            delivery_mode=2,
            loop=None,
            serializer=None,
    ):
        """There must be at least these members of the connection_parameters dict:
            'connection_parameters': {
                'login': '',
                'password': '',
                'host': '',
                'port': '',
                'virtualhost': '',
            },


        :param connection_parameters: Dict with connection parameters. See above for its format.
        :return: EventsQueueClient instance.
        """

        # Can not pass empty password when connecting. Must remove the field completely.
        if not connection_parameters.get('password', ''):
            connection_parameters.pop('password', None)

        self._connection_parameters = connection_parameters
        self._transport = None
        self._protocol = None
        self._channel: Channel = None
        self.channel_name = channel_name
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._delivery_mode = delivery_mode
        self._connection_timeout = connection_timeout
        self._is_connecting = False
        self._connected = False
        self._connection_guid = str(uuid4())
        self._known_queues = {}
        self._routing = {}
        self.loop = loop
        self._serializer = serializer or JSONSerializer()
        self._response_queue = None
        self.requests = WeakValueDictionary()

        # TODO move to config
        self.retry_exchange = 'retries'
        self.default_command_retries = 100
        self.default_timeout = 12

    @property
    def routing(self):
        return self._routing

    @property
    def connected(self):
        return self._connected or (self._channel is not None and self._channel.is_open)

    @property
    def is_connecting(self):
        return self._is_connecting

    @property
    def channel(self):
        return self._channel

    @property
    def transport(self):
        return self._transport

    @property
    def protocol(self) -> aioamqp.AmqpProtocol:
        return self._protocol

    async def connect(self):
        """
        Create new asynchronous connection to the RabbitMQ instance.
        This will connect, declare exchange and bind itself to the configured queue.

        After that, client is ready to publish or consume messages.

        :return: Does not return anything.
        """
        if self.connected or self.is_connecting:
            return

        self._is_connecting = True
        try:
            logger.info(
                'Connecting to RabbitMQ on %s/%s...',
                self._connection_parameters.get('host'),
                self._connection_parameters.get('virtualhost')
            )
            self._transport, self._protocol = await asyncio.wait_for(
                aioamqp.connect(
                    loop=self.loop,
                    on_error=self.callback_error,
                    **self._connection_parameters
                ),
                self._connection_timeout
            )
            self._channel = await self._protocol.channel()
            self._connected = True
        except (AmqpClosedConnection,
                ChannelClosed,
                ConnectionRefusedError,
                asyncio.CancelledError) as err:

            self._is_connecting = False
            logger.exception('RabbitMQ connection error')
            self.callback_error(err)

        self._is_connecting = False
        return self._transport, self._protocol

    def callback_error(self, exception):
        raise exceptions.StreamConnectionError('Error initializing RabbitMQ connection') \
            from exception

    async def command(self, data, routing_key):
        if not self.connected:
            logger.warning('Attempted to send message while not connected')
            return

        body = self._serializer.serialize(Response(body=data))

        # TODO use context!
        default_properties = {
            'delivery_mode': self._delivery_mode,
            'headers': {
                Headers.correlation_id: self.generate_correlation_id(),
                Headers.retry_attempts_max: str(self.default_command_retries),
                Headers.expired_at: str(self.default_timeout),
            }
        }

        logger.info('sbus <~~~ %s: %s', routing_key, body[:256])

        await self.channel.publish(
            payload=body,
            exchange_name=self._exchange_name,
            routing_key=routing_key,
            properties=default_properties,
        )

    async def on(self, subscriber: AbstractSubscriber):
        """Subscribe to the queue consuming."""
        await self.declare_queue(subscriber.name, subscriber.durable)

        for routing_key in subscriber.routing_keys:
            await self.bind_to_queue(routing_key, subscriber.name)

        for key in subscriber.routing_keys:
            if subscriber == self._routing.get(key):
                logger.warning(
                    'Subscriber "%s" already received routing_key "%s".',
                    subscriber,
                    key
                )
                break
            self._routing[key] = subscriber

        logger.info('Starting consumption for %s', subscriber)

        await self._channel.basic_qos(prefetch_count=subscriber.prefetch_count)
        await self._channel.basic_consume(
            callback=self._on_message,
            consumer_tag=subscriber.consumer_tag(),
            queue_name=subscriber.name,
        )

        self._add_to_known_queues(subscriber)

    async def request(self, data, routing_key, response_class):
        """RPC call method

        :param data: query data
        :param routing_key: routing key
        :param response_class: response class
        """
        if not self.connected:
            logger.warning('Attempted to send message while not connected')
            return

        corr_id = self.generate_correlation_id()
        callback_queue = await self._get_response_queue()

        body = self._serializer.serialize(Response(body=data))
        waiter_future = asyncio.Future()

        self.requests[corr_id] = waiter_future

        async def waiter_callback(ch, body, method, props):
            if props.correlation_id in self.requests:
                waiter = self.requests.get(props.correlation_id)
                waiter.set_result(body)
            else:
                logger.warning('There is no callback for request with correlation id %s',
                               props.correlation_id)

        logger.debug('Start callback consuming on queue: %s', callback_queue)
        await self.channel.basic_consume(callback=waiter_callback, queue_name=callback_queue,
                                         no_ack=True)

        logger.info('sbus ~~~> %s: %s', routing_key, body[:256])

        await self.channel.publish(
            exchange_name=self._exchange_name,
            routing_key=routing_key,
            properties={
                'reply_to': callback_queue,
                'correlation_id': corr_id,
                'headers': {
                    Headers.correlation_id: self.generate_correlation_id(),
                    Headers.retry_attempts_max: self.default_command_retries,
                    Headers.expired_at: str(self.default_timeout),
                }
            },
            payload=body
        )

        # TODO parametrize timeout for request 14.05.2018
        response = await asyncio.wait_for(self.requests[corr_id], self.default_timeout)
        logger.info('sbus <~~~ %s: %s', routing_key, response[:256])
        return self._serializer.deserialize(response, response_class)

    async def declare_queue(self, queue_name, durable=True):
        logger.info('Declaring queue...')
        queue_declaration = await self._channel.queue_declare(queue_name, durable=durable)
        queue_name = queue_declaration.get('queue')
        logger.info('Declared queue "%s"', queue_name)

    async def declare_exchange(self, exchange_name, type_name='direct', durable=True):
        logger.info('Declaring exchange...')
        await self._channel.exchange_declare(exchange_name, type_name=type_name, durable=durable)
        logger.info('Declared exchange "%s" with type "%s" and durable=%s', exchange_name,
                    type_name, durable)

    async def bind_to_queue(self, routing_key, queue_name):
        """Bind to queue with specified routing key."""
        logger.info('Binding exchange="%s", queue="%s", key="%s"',
                    self._exchange_name, queue_name, routing_key)

        result = await self._channel.queue_bind(
            exchange_name=self._exchange_name,
            queue_name=queue_name,
            routing_key=routing_key,
        )
        return result

    async def _on_message(self, channel, body, envelope, properties):
        """Subscriber messages handler."""
        try:
            logger.debug('Processing message correlation_id=%s', properties.correlation_id)
            logger.info('sbus ~~~> %s: %s', envelope.routing_key, body[:256])

            subscriber: AbstractSubscriber = self._routing.get(envelope.routing_key)
            if not subscriber:
                logger.debug('No route for message with key "%s"', envelope.routing_key)
                return

            context = Context(message_id=properties.message_id, routing_key=envelope.routing_key)

            response_body = await subscriber.on_message(
                body, envelope.routing_key, context, self._serializer
            )

            if not properties.reply_to:
                await self.ack(envelope.delivery_tag)
                return

            else:
                # TODO log correlation id as additional field
                # TODO move to config body slice
                logger.info('sbus resp <~~~ %s: %s', envelope.routing_key, response_body[:256])

                await self.channel.publish(
                    exchange_name='',
                    routing_key=properties.reply_to,
                    properties={
                        'correlation_id': properties.correlation_id,
                    },
                    payload=response_body
                )

            await self.ack(envelope.delivery_tag)

        except exceptions.UnrecoverableExceptionBase as error:
            await self._handle_unrecoverable_exception(envelope, properties, error)

        # any exception which not marked as unrecoverable will be retried
        except Exception as error:
            await self._handle_recoverable_exception(envelope, body, properties, error)

    async def _handle_unrecoverable_exception(self, envelope: Envelope, properties: Properties,
                                              error: exceptions.UnrecoverableExceptionBase):
        logger.error(
            'Cannot process message "%s" in "%s" because of %s.',
            envelope.delivery_tag,
            envelope.routing_key,
            error
        )
        await self._on_fail(envelope, properties, error)

    async def _handle_recoverable_exception(self, envelope: Envelope,
                                            payload: bytes, properties: Properties,
                                            error: Exception):
        logger.exception(
            'Cannot process message "%s" in "%s"!. Because of %s.',
            envelope.delivery_tag,
            envelope.routing_key,
            error
        )

        headers: typing.Dict = properties.headers or {}

        # TODO aioamqp can not publish x-death header because it's type is list
        # right now i do not need lists in headers but this should be workaround somehow
        headers.pop('x-death', None)

        attempts_max: typing.AnyStr = smart_int(headers.get(Headers.retry_attempts_max))
        attempts_nr: typing.AnyStr = smart_int(headers.get(Headers.retry_attempt_nr), 1)

        if attempts_max and attempts_max > attempts_nr:
            headers[Headers.retry_attempt_nr] = str(attempts_nr + 1)

            backoff: int = int(math.pow(2, min(attempts_nr - 1, 7))) * 1000

            new_properties = {
                'delivery_mode': 2,
                'headers': headers,
                'expiration': str(backoff),
            }

            expires_at = headers.get(Headers.expired_at)
            if expires_at and float(expires_at) <= (time.time() * 1000 + backoff):
                logger.exception(
                    'timeout %s. Message will be expired %s, '
                    'don\'t retry it! Correlation id: %s',
                    envelope.routing_key,
                    expires_at,
                    headers.get(Headers.correlation_id),
                )
                await self._on_fail(envelope, properties, error)

            else:
                try:
                    logger.exception(
                        'Routing key: %s. Retry attempt %s after %s millis...',
                        envelope.routing_key, attempts_nr, backoff
                    )
                    await self.channel.publish(
                        exchange_name=self.retry_exchange,
                        routing_key=envelope.routing_key,
                        properties=new_properties,
                        payload=payload
                    )
                except Exception:  # noqa
                    logger.exception('Error on publish retry message for %s',
                                     envelope.routing_key)
                finally:
                    await self._on_fail(envelope, properties, error)
        else:
            logger.exception(
                'max retries (%s) reached %s, '
                'don\'t retry it! Correlation id: %s',
                attempts_max,
                envelope.routing_key,
                headers.get(Headers.correlation_id),
            )

            await self.nack(envelope.delivery_tag)

    async def _on_fail(self, envelope: Envelope, properties: Properties, error: Exception):
        """ Call when we have no chance to process message."""
        try:
            if properties.reply_to:
                response = self._serializer.serialize(
                    Response(
                        status=getattr(error, 'status', 500),
                        body=ResponseBody(error=str(error)).dict()
                    )
                )

                logger.info('sbus resp <~~~ %s: %s', envelope.routing_key, response[:256])

                await self.channel.publish(
                    exchange_name='',
                    routing_key=properties.reply_to,
                    properties={'correlation_id': properties.correlation_id},
                    payload=response
                )

        finally:
            await self.nack(envelope.delivery_tag)

    def _add_to_known_queues(self, subscriber: AbstractSubscriber):
        self._known_queues[subscriber.name] = subscriber

    @staticmethod
    def generate_correlation_id():
        corr_id = str(uuid4())
        logger.debug('Generated correlation_id %s', corr_id)
        return corr_id

    async def _get_response_queue(self):
        if not self._response_queue:
            logger.info('Declaring rpc queue...')
            self._request_queue = await self.channel.queue_declare(exclusive=True)
            logger.info('Declared rpc queue "%s"', self._request_queue.get('queue'))
        return self._request_queue.get('queue')

    async def ack(self, delivery_tag):
        await self._channel.basic_client_ack(delivery_tag)
        logger.debug('Message "%s" was acked.', delivery_tag)

    async def nack(self, delivery_tag):
        await self._channel.basic_client_nack(delivery_tag, requeue=False)
        logger.debug('Message "%s" was negative acked.', delivery_tag)

    async def destroy(self, *args, **kwargs):
        # cancel consuming
        for subscriber in self._known_queues.values():
            logger.info('Cancelling consumption for %s', subscriber)
            await self._channel.basic_cancel(subscriber.consumer_tag())

        if self._channel and self._channel.is_open:
            await self._channel.close()

        if self._protocol is not None and self._protocol.state != CLOSED:
            # wait till all heartbeats is stopped
            wait = kwargs.get('wait', False)

            try:
                # close using the `AMQP` protocol
                self.protocol.connection_lost(exc=None)
                await self.protocol.close(no_wait=not wait, timeout=10)

                # ensure the socket is closed.
                self.transport.close()
            except (aioamqp.AmqpClosedConnection, exceptions.StreamConnectionError):
                logger.info('Graceful connection shutdown detected')

        self._connected = False
        self._is_connecting = False



"""
Abstract and bases for building messaging stuff.
"""

from abc import ABCMeta, abstractmethod


class AbstractTransport(metaclass=ABCMeta):

    @abstractmethod
    async def connect(self):
        pass

    @property
    @abstractmethod
    def connected(self):
        pass

    @property
    @abstractmethod
    def transport(self):
        pass

    @property
    @abstractmethod
    def protocol(self):
        pass

    @abstractmethod
    async def command(self, data, routing_key):
        pass

    @abstractmethod
    async def on(self, subscriber):
        pass

    @abstractmethod
    async def request(self, data, routing_key, response_class):
        pass

    @abstractmethod
    async def destroy(self):
        pass



import json
from unittest import mock

import pytest

from sbus.sbus import BackOff, Sbus, SbusWatchDog
from sbus.transports import SUPPORTED_TRANSPORTS, AMQPTransport
from sbus.utils import create_transport
from tests.conftest import Context  # noqa
from tests.utils import make_mocked_coro


@pytest.fixture
def transport_config(connection_parameters, exchange_name, exchange_type, worker_name):
    return {
        'transports': {
            'amqp': {
                'connection_parameters': connection_parameters,
                'exchange_name': exchange_name,
                'exchange_type': exchange_type,
                'channel_name': worker_name,
                'delivery_mode': 2,
                'connection_timeout': 10,
            }
        },
        'active_transport': 'amqp',
    }


@pytest.fixture
def transport(transport_config, event_loop):
    return create_transport(transport_config, event_loop)


class TestServiceBus:

    @pytest.fixture
    def sbus(self, event_loop, transport_config) -> Sbus:
        active_transport = transport_config['active_transport']
        active_transport_setup = transport_config['transports'][active_transport]
        transport_class = SUPPORTED_TRANSPORTS[active_transport]
        transport = transport_class(**active_transport_setup)
        return Sbus(transport, loop=event_loop)

    def test_service_bus_init(self, sbus: Sbus):
        assert isinstance(sbus.transport, AMQPTransport)
        assert isinstance(sbus.subscribers, set)

    @pytest.mark.asyncio
    async def test_connect(self, sbus: Sbus):
        mocked_connect = make_mocked_coro()
        sbus.transport.connect = mocked_connect
        await sbus.connect()
        assert mocked_connect.called

    def test_connected(self, sbus: Sbus):
        mocker = mock.patch('tests.test_service_bus.AMQPTransport.connected',
                            new=mock.PropertyMock(return_value=True))
        with mocker as mocked_connected:
            sbus.transport.connected = mocked_connected
            result = sbus.connected
            assert result
            assert mocked_connected.called

    @pytest.mark.asyncio
    async def test_publish(self, sbus: Sbus):
        mocked_publish = make_mocked_coro()
        sbus.transport.command = mocked_publish
        await sbus.command('message', 'routing_key')
        mocked_publish.assert_called_once()

    @pytest.fixture
    def callback(self):
        def func(ch, method, props, body):
            pass
        return func

    @pytest.mark.asyncio
    async def test_dequeue(self, sbus: Sbus, subscriber):
        mocked_consume_queue = make_mocked_coro()
        sbus.transport.on = mocked_consume_queue
        with mock.patch('tests.test_service_bus.AMQPTransport.connected',
                        new=mock.PropertyMock(return_value=True)):
            await sbus.on(subscriber)
            assert mocked_consume_queue.called

    @pytest.mark.asyncio
    async def test_stop(self, sbus: Sbus):
        mocked_destroy = make_mocked_coro()
        with mock.patch('tests.test_service_bus.AMQPTransport.destroy',
                        new=mocked_destroy):
            await sbus.stop()
            assert mocked_destroy.called


class TestServiceBusWatchDog:

    def test_class_interface(self, transport_config, worker_name, connection_parameters):
        loop = mock.Mock()

        service_bus_factory = Sbus.create_factory(transport_config, loop)
        watch_dog = SbusWatchDog(service_bus_factory)

        assert watch_dog.sbus.loop == loop
        assert watch_dog.sbus.transport.channel_name == worker_name
        assert watch_dog.sbus.transport._connection_parameters == connection_parameters

    @pytest.mark.asyncio
    async def test_repair_connection(self, ctx: 'Context', transport_config, event_loop,
                                     echo_subscriber, amqp_client, test_topic):

        service_bus_factory = Sbus.create_factory(transport_config, event_loop)
        backoff = BackOff(max_timeout=10)
        watch_dog = SbusWatchDog(service_bus_factory, backoff=backoff)

        on_message_queue = ctx.create_queue(size=1, timeout=10)

        async def on_message(data, routing_key, context_, serializer_):
            on_message_queue.put_nowait((data, routing_key, context_, serializer_))

        echo_subscriber.on_message = on_message

        await watch_dog.sbus.connect()
        await watch_dog.sbus.on(echo_subscriber)

        assert watch_dog.sbus.clone_subscribers() == {echo_subscriber}

        published_message = {'hello': 'world'}

        await amqp_client.command(published_message, test_topic)

        r_message, r_key, r_context, r_serializer = await on_message_queue.get()
        assert (json.loads(r_message).get('body'), r_key) == (published_message, test_topic)

        # TODO: takes a lot of time to complete
        # await watch_dog.sbus.transport.protocol.close(no_wait=False)

        # we must wait until connection is fully closed
        await watch_dog.repair()

        await amqp_client.command(published_message, test_topic)

        r_message, r_key, r_context, r_serializer = await on_message_queue.get()
        assert (json.loads(r_message).get('body'), r_key) == (published_message, test_topic)

        await watch_dog.sbus.stop()



"""
Exceptions, related to messaging.
"""


class UnrecoverableExceptionBase(Exception):
    """Exception for negative ack to service bus"""


class StreamConnectionError(Exception):
    """Error connecting transport"""


class ConsumerError(StreamConnectionError):
    """Some error in message consumer"""


class PublisherError(StreamConnectionError):
    """Problem with message publisher"""


class SerializationError(UnrecoverableExceptionBase):
    """Error serializing or deserializing data"""


class StreamInterrupt(BaseException):
    """Error for stop StreamWorker loop gracefully."""


class RecoverableErrorBase(UnrecoverableExceptionBase):
    pass


class TooManyRequestError(UnrecoverableExceptionBase):
    status = 429


class BadRequestError(UnrecoverableExceptionBase):
    status = 400


class UnauthorizedError(UnrecoverableExceptionBase):
    status = 401


class ForbiddenError(UnrecoverableExceptionBase):
    status = 403


class NotFoundError(UnrecoverableExceptionBase):
    status = 404


class MethodNotAllowedError(UnrecoverableExceptionBase):
    status = 405


class UnrecoverableError(UnrecoverableExceptionBase):
    status = 456


class ServiceUnavailableError(UnrecoverableExceptionBase):
    status = 503


class InternalServerError(UnrecoverableExceptionBase):
    status = 500



#!/usr/bin/env python
import os

from setuptools import find_packages, setup

from sbus import __version__

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='sbus',
    version=__version__,
    author='Andrey Zdanchuk',
    author_email='dyuuus@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    license='BSD License',
    description='Pub/sub abstraction for aioamqp',
    install_requires=[
        'aioamqp==0.10.0',
        'pydantic==0.6.3'
    ],
    tests_require=[
        'flake8-builtins',
        'flake8-quotes',
        'flake8-pep3101',
        'flake8-comprehensions',
        'flake8-blind-except',
        'pep8-naming',

        'pytest-cov',
        'pytest-isort',

        'pytest-asyncio',
        'flake8',
        'pytest-flake8',
        'pytest'
    ],
    setup_requires=['pytest-runner'],
    url='https://github.com/dyuuus/sbus'
)



import json

import pytest
from aioamqp.channel import Channel

from sbus.exceptions import StreamConnectionError
from sbus.transports import AMQPTransport
from tests.utils import make_mocked_coro


class TestAMQPTransport:
    @pytest.fixture
    def setup_queue(self, event_loop, amqp_client, exchange_name, queue_name,
                    test_topic):
        event_loop.run_until_complete(amqp_client.declare_queue(queue_name))
        event_loop.run_until_complete(amqp_client.bind_to_queue(test_topic, queue_name))

    @pytest.fixture
    def not_connected_client(self, connection_parameters, exchange_name, exchange_type,
                             worker_name, event_loop):
        return AMQPTransport(connection_parameters, exchange_name, exchange_type, worker_name,
                             loop=event_loop)

    def test_amqp_client_init_and_connect(
        self, amqp_client, connection_parameters, exchange_name, exchange_type, worker_name,
        event_loop,
    ):
        """Test initial state after connect."""
        assert amqp_client._connection_parameters == connection_parameters
        assert amqp_client._transport
        assert amqp_client._protocol
        assert amqp_client._channel
        assert amqp_client._serializer
        assert amqp_client.channel_name == worker_name
        assert amqp_client._exchange_name == exchange_name
        assert amqp_client._exchange_type == exchange_type
        assert amqp_client._delivery_mode == 2
        assert amqp_client._connection_timeout == 3
        assert not amqp_client._is_connecting
        assert amqp_client._connection_guid
        assert amqp_client._known_queues == {}
        assert amqp_client._routing == {}
        assert amqp_client.loop == event_loop
        assert amqp_client.routing == {}
        assert amqp_client.connected is True
        assert not amqp_client.is_connecting
        assert isinstance(amqp_client.channel, Channel)
        assert amqp_client.transport
        assert amqp_client.protocol

    @pytest.mark.asyncio
    async def test_callback_error(self, ctx, not_connected_client):
        not_connected_client.connect = make_mocked_coro(raise_exception=StreamConnectionError)
        with pytest.raises(StreamConnectionError):
            await not_connected_client.connect()

    @pytest.mark.asyncio
    async def test_publish(self, ctx, setup_queue, amqp_client,
                           echo_subscriber, test_topic):
        assert amqp_client.connected
        message = {'test': 'hello'}
        await amqp_client.command(message, test_topic)

        on_message_queue = ctx.create_queue(size=1, timeout=3)

        async def on_message(data, routing_key, context_, serializer_):
            on_message_queue.put_nowait((data, routing_key, context_, serializer_))

        echo_subscriber.on_message = on_message
        await amqp_client.on(echo_subscriber)
        r_message, r_key, r_context, r_serializer = await on_message_queue.get()
        assert (json.loads(r_message).get('body'), r_key) == (message, test_topic)

    @pytest.mark.asyncio
    async def test_client_destroy(self, ctx, not_connected_client):
        """
        Test client destroyable.
        """
        await not_connected_client.connect()
        await not_connected_client.destroy()

    @pytest.mark.asyncio
    async def test_consume_queue(self, ctx, setup_queue, amqp_client, subscriber):
        """
        Test that can subscribe.
        """
        await amqp_client.on(subscriber)

        assert subscriber.routing_keys == tuple(amqp_client._routing.keys())
        assert subscriber in list(amqp_client._routing.values())
        assert subscriber.name in amqp_client._known_queues




class Router:
    """Register your consumers here."""
    _routing = {}

    def __init__(self, subscribers=None):
        if subscribers is not None:
            self.routing.update(subscribers)

    @property
    def routing(self):
        return self._routing

    def register(self, *routing_keys):
        def deco(func):
            for key in routing_keys:
                self.routing.update({key: func})
            return func

        return deco

    def get_handler(self, routing_key):
        return self.routing.get(routing_key)
